# coding: utf-8

import os.path
import cv2
import cvutils
import utils
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QGraphicsView, QFileDialog, QMessageBox, QUndoStack
from PyQt5.QtCore import QObject, pyqtSignal, QRectF, QEvent, QCoreApplication
from PyQt5.QtGui import QPixmap, QImage, QPainter, QColor, QPen, QBrush, QKeyEvent
from undo import UndoPen, UndoMask


class PenLine:
    '''ペンの色と線の座標'''

    def __init__(self, penColor, penSize):
        self.penColor = penColor
        self.penSize = float(penSize)
        self.points = []

    def penColor(self):
        return self.penColor

    def setPoint(self, point):
        self.points = [point.toPoint()]

    def addPoint(self, point):
        self.points.append(point.toPoint())

    def getPoints(self):
        return self.points

    def setPoints(self, points):
        self.points = points

    def drawPolyline(self, painter):
        if len(self.points) < 2:
            return
        pen = QPen(QBrush(self.penColor), self.penSize)
        pen.setCapStyle(Qt.RoundCap)
        pen.setJoinStyle(Qt.RoundJoin)
        painter.setPen(pen)
        painter.drawPolyline(*self.points)


class EventHandler(QObject):
    '''MainWindowのイベントハンドラとモデル'''

    # RGBのしきい値が変化した
    redValueChanged = pyqtSignal('int')
    greenValueChanged = pyqtSignal('int')
    blueValueChanged = pyqtSignal('int')
    # スライダー類の有効/無効が変化した
    enableSlidersChanged = pyqtSignal('bool')
    # マスクの不透明度が変化した(intはスライダー用)
    maskOpacityChangedFloat = pyqtSignal('float')
    maskOpacityChangedInt = pyqtSignal('int')
    # ペンのサイズが変化した
    penSizeChanged = pyqtSignal('int')
    # 前フレームマスクの所持/不所持が変化した
    hasPreviousFrameChanged = pyqtSignal('bool')

    def __init__(self, window, parent=None):
        super().__init__(parent)
        self.window = window
        # 原画像のパス
        self.srcPath = None
        # 同時に動かす, 有効
        self.sliderSync = False
        self.enableSliders = True
        # RGBしきい値
        self.redValue = 255
        self.greenValue = 120
        self.blueValue = 255
        # ペンの色
        self.penColor = None
        self.penSize = 2
        # マスクのα
        self.maskOpacity = 0.5
        # 前フレームあり/なし
        self.hasPreviousFrame = False
        #
        self.pressPoint = None
        self.currentLine = None
        self.penLines = []
        self.undoStack = QUndoStack()
        ui = self.window.ui
        self.valueNames = {
            # スライダーのsenderと名前
            ui.redValue: 'redValue',
            ui.greenValue: 'greenValue',
            ui.blueValue: 'blueValue',
        }

    def setSliderSync(self, checked):
        '''
        同時に動かすが変化したときのハンドラ
        チェックが入った場合は緑の値に合わせる
        '''
        self.sliderSync = checked
        if checked:
            self.redValue = self.blueValue = self.greenValue
            self.redValueChanged.emit(self.redValue)
            self.blueValueChanged.emit(self.blueValue)

    def setEnableSliders(self, enabled):
        self.enableSliders = enabled
        self.enableSlidersChanged.emit(enabled)

    def setValue(self, value):
        '''スライダー/スピンボックスが変化したときのハンドラ'''
        name = self.valueNames.get(self.sender())
        if name is None:
            return
        names = [name]
        if self.sliderSync:
            names = self.valueNames.values()
        for name in names:
            setattr(self, name, value)
            changed = '%sChanged' % name
            getattr(self, changed).emit(value)
        self.plantSegmentation()

    def setPenColor(self, index):
        '''colorComboBoxのハンドラ'''
        self.penColor = self.window.ui.colorComboBox.currentData()

    def setPenSize(self, size):
        self.penSize = size
        self.penSizeChanged.emit(size)

    def setMaskOpacityInt(self, opacity):
        '''マスクのαが変化したときのハンドラ'''
        self.maskOpacity = float(opacity) / 100
        self.maskOpacityChangedFloat.emit(self.maskOpacity)
        self.maskOpacityChangedInt.emit(opacity)

    def plantSegmentation(self):
        '''
        プラントセグメンテーション
        原画像を3原色に分解し、それぞれのチャネルに対して2値化をおこなう。
        2値化されたチャネルをorで合成し、メインウィンドウのマスク画像を置き換える
        データは常にRGBで、OpenCV標準のBGRではない
        '''
        mats = cvutils.decompose(self.window.getSrcPixmap())
        ret, mats[0] = cv2.threshold(mats[0], self.redValue, 255, cv2.THRESH_BINARY_INV)
        ret, mats[1] = cv2.threshold(mats[1], self.greenValue, 255, cv2.THRESH_BINARY_INV)
        ret, mats[2] = cv2.threshold(mats[2], self.blueValue, 255, cv2.THRESH_BINARY_INV)
        mat = cv2.bitwise_or(cv2.bitwise_or(mats[0], mats[1]), mats[2])
        self.window.setMaskPixmap(cvutils.compose((mat, mat, mat)))

    def isPanMode(self):
        return self.window.ui.graphicsView.dragMode() != QGraphicsView.NoDrag

    def mousePressEvent(self, event):
        if not event.buttons() & Qt.LeftButton:
            return
        if self.isPanMode():
            self.pressPoint = event.scenePos()
        else:
            maskPixmap = self.window.getMaskPixmap()
            point = event.scenePos().toPoint()
            if cvutils.pixmapPixel(maskPixmap, point) == 0 or self.isWhiteOrBlack(self.penColor):
                # クリックした点が黒だった
                # もしくはペンが白か黒だった
                # ペンツール
                self.currentLine = PenLine(self.penColor, self.penSize)
                self.currentLine.setPoint(event.scenePos())
                self.window.getPenItem().update()
            else:
                # バケツツール
                if self.isBlack(self.penColor):
                    return
                pixmap = cvutils.floodFill(maskPixmap, point, self.penColor)
                if pixmap is not None:
                    self.undoStack.push(UndoMask(self.window.getMaskItem(), maskPixmap, pixmap))

    def mouseMoveEvent(self, event):
        if not event.buttons() & Qt.LeftButton:
            return
        if self.isPanMode():
            if self.currentLine is not None:
                # ペンで書いてる最中にaltキーを押した
                self.currentLine = None
                self.window.getPenItem().update()
                return
            if self.pressPoint is None:
                return
            #http://stackoverflow.com/questions/4753681/how-to-pan-images-in-qgraphicsview
            view = self.window.ui.graphicsView
            hScroll = view.horizontalScrollBar()
            vScroll = view.verticalScrollBar()
            delta = (self.pressPoint - event.scenePos()).toPoint()
            hScroll.setValue(hScroll.value() + delta.x())
            vScroll.setValue(vScroll.value() + delta.y())
        elif self.currentLine is not None:
            self.currentLine.addPoint(event.scenePos())
            self.window.getPenItem().update()

    def mouseReleaseEvent(self, event):
        if self.currentLine is None:
            return
        self.currentLine.addPoint(event.scenePos())
        if self.isWhiteOrBlack(self.penColor):
            # ペンの色が白か黒ならマスクを修正する
            oldMask = self.window.getMaskPixmap()
            mask = oldMask.copy()
            painter = QPainter(mask)
            line = PenLine(self.penColor, self.penSize)
            line.setPoints(self.currentLine.getPoints())
            line.drawPolyline(painter)
            painter.end()
            self.currentLine = None
            self.window.getPenItem().update()
            self.window.setMaskPixmap(mask)
            self.undoStack.push(UndoMask(self.window.getMaskItem(), oldMask, mask))
            return
        # 開始と終了が黒画素でない場合は不完全な入力なのでペンの更新はおこなわない
        # という仕様
        image = self.window.getMaskPixmap().toImage()
        points = self.currentLine.getPoints()
        startIsBlack = self.isBlack(image.pixelColor(points[0]))
        endIsBlack = self.isBlack(image.pixelColor(points[-1]))
        if startIsBlack and endIsBlack:
            self.penLines.append(self.currentLine)
        self.currentLine = None
        self.window.getPenItem().update()
        self.undoStack.push(UndoPen(self.window.getPenItem(), self.penLines))

    def keyPressEvent(self, event):
        if event.modifiers() & Qt.ControlModifier:
            self.window.ui.graphicsView.setDragMode(QGraphicsView.ScrollHandDrag)

    def keyReleaseEvent(self, event):
        self.window.ui.graphicsView.setDragMode(QGraphicsView.NoDrag)

    def wheelEvent(self, event):
        if not self.isPanMode():
            return
        view = self.window.ui.graphicsView
        view.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)
        scale = 1.1
        if 0 < event.delta():
            view.scale(scale, scale)
        else:
            view.scale(1/scale, 1/scale)

    def resizeEvent(self, event):
        '''
        リサイズイベント
        モデルの初期値をビューに設定する
        '''
        self.redValueChanged.emit(self.redValue)
        self.greenValueChanged.emit(self.greenValue)
        self.blueValueChanged.emit(self.blueValue)
        self.maskOpacityChangedFloat.emit(self.maskOpacity)
        self.maskOpacityChangedInt.emit(int(self.maskOpacity * 100))
        self.penSizeChanged.emit(self.penSize)
        self.hasPreviousFrameChanged.emit(self.hasPreviousFrame)
        # 色は無理なのでビューに合わせる
        colorComboBox = self.window.ui.colorComboBox
        currentIndex = colorComboBox.currentIndex()
        self.penColor = colorComboBox.itemData(currentIndex)

    def paintEvent(self, event):
        print('paintEvent')

    def paint(self, painter, item, widget):
        if self.currentLine is not None:
            self.currentLine.drawPolyline(painter)
        for line in self.penLines:
            line.drawPolyline(painter)

    def isBlack(self, color):
        return color.red() == 0 and color.green() == 0 and color.blue() == 0

    def isWhiteOrBlack(self, color):
        r = color.red()
        g = color.green()
        b = color.blue()
        return (r == g and r == b) and (r == 0 or r == 255)

    def open(self, fname):
        '''原画像を開く'''
        dir = '.'
        if self.srcPath is not None:
            dir = os.path.dirname(self.srcPath)
        fname = QFileDialog.getOpenFileName(self.window, 'Open file', dir, 'Images (*rgb*.png)')[0]
        if not os.path.exists(fname):
            return
        # 原画
        pixmap = QPixmap(fname)
        size = pixmap.size()
        blackPixmap = QPixmap(size)
        blackPixmap.fill(QColor(0, 0, 0))
        self.window.getSrcItem().setPixmap(pixmap)
        # マスク
        maskPath = self.getMaskPath(fname)
        if maskPath is not None and os.path.exists(maskPath):
            # すでにマスクがある場合はそれを読み込む, スライダー類は無効にする
            self.window.getMaskItem().setPixmap(QPixmap(maskPath))
            self.setEnableSliders(False)
        else:
            self.window.getMaskItem().setPixmap(blackPixmap.copy())
            self.setEnableSliders(True)
            self.plantSegmentation()
        # 前フレーム
        #previousPath = self.getPreviousPath(fname)
        previousPath = None
        if previousPath is not None and os.path.exists(previousPath):
            self.window.getPreviousItem().setPixmap(QPixmap(previousPath))
            self.setHasPreviousFrame(True)
        else:
            self.window.getPreviousItem().setPixmap(blackPixmap.copy())
            self.setHasPreviousFrame(False)
        # ペン
        self.window.getPenItem().setPixmap(blackPixmap.copy())
        self.window.scene.setSceneRect(QRectF(0, 0, size.width(), size.height()))
        self.srcPath = os.path.abspath(fname)
        self.window.setWindowTitle(os.path.basename(fname))
        self.currentLine = None
        self.penLines = []
        self.undoStack.clear()
        # XXX: ショートカットキーからだとリリースイベントが飛んでこないので
        #QCoreApplication.sendEvent(self.window.scene, QKeyEvent(QEvent.KeyPress, 0, Qt.NoModifier))
        QCoreApplication.sendEvent(self.window.scene, QKeyEvent(QEvent.KeyRelease, 0, Qt.NoModifier))

    def getMaskPath(self, srcPath):
        '''原画ファイル名からマスクファイル名を求める'''
        if srcPath is None:
            return None
        srcName = os.path.basename(srcPath)
        print(srcName)
        if (srcName.index('rgb') == -1):
            raise '原画ファイル名が "rgb" を含んでいません'
        maskName = 'label'.join(srcName.rsplit('rgb', 1))
        return os.path.join(os.path.dirname(srcPath), maskName)

    def getPreviousPath(self, srcPath):
        '''原画ファイル名から前フレームのマスクファイル名を求める'''
        return self.getMaskPath(utils.getPreviousFrame(srcPath))

    def setHasPreviousFrame(self, hasPreviousFrame):
        self.hasPreviousFrame = hasPreviousFrame
        self.hasPreviousFrameChanged.emit(hasPreviousFrame)

    def showPreviousFrame(self, show):
        if show:
            self.window.getPreviousItem().setOpacity(1)
        else:
            self.window.getPreviousItem().setOpacity(0)

    def save(self):
        '''マスク画像を保存'''
        maskPath = self.getMaskPath(self.srcPath)
        if maskPath is None:
            return
        if os.path.exists(maskPath):
            msgbox = QMessageBox(self.window)
            msgbox.setIcon(QMessageBox.Warning)
            msgbox.setText('マスクファイルが存在します')
            msgbox.setInformativeText('すでにマスクファイル存在しています。上書きしますか?')
            msgbox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            if msgbox.exec() != QMessageBox.Yes:
                return
        self.applyPenLayer()
        self.window.getMaskPixmap().save(maskPath)
        self.window.ui.statusbar.showMessage('%sを保存しました' % os.path.basename(maskPath))

    def merge(self):
        '''ペンレイヤーとマスクレイヤーの結合'''
        self.applyPenLayer()

    def undo(self):
        '''アンドゥ'''
        self.undoStack.undo()

    def redo(self):
        self.undoStack.redo()

    def applyPenLayer(self):
        '''
        ペンレイヤーの線をマスク画像に反映させる
        その際、分割線のはみ出ている部分は黒にする
        '''
        whiteColor = QColor(255, 255, 255)
        maskPixmap = self.window.getMaskPixmap()
        alphaPixmap = maskPixmap.copy()
        alphaPixmap.setMask(maskPixmap.createMaskFromColor(whiteColor))
        dstPixmap = QPixmap(maskPixmap.size())

        # 1. 最初に白で埋めておく
        # 2. 上から分割線を描画する
        # 3. さらに上からマスク画像の白以外(==黒)を重ねる
        dstPixmap.fill(whiteColor)
        painter = QPainter(dstPixmap)
        for line in self.penLines:
            line.drawPolyline(painter)
        painter.drawPixmap(dstPixmap.rect(), alphaPixmap, alphaPixmap.rect())
        painter.end()
        self.window.setMaskPixmap(dstPixmap)
        self.undoStack.push(UndoMask(self.window.getMaskItem(), maskPixmap, dstPixmap))
        self.window.ui.graphicsView.setFocus()
