# coding: utf-8

import sys
import os
import codecs
import csv
from PyQt5.QtWidgets import QMainWindow, QGraphicsScene, QGraphicsPixmapItem
from PyQt5.QtGui import QPixmap, QColor, QIcon
from Ui_MainWindow import Ui_MainWindow
from EventHandler import EventHandler


class MainWindow(QMainWindow):
    '''
    メインウィンドウ

    MainWindowはビューの保持とシグナル/スロット/イベントハンドラの初期設定が主な目的である
    ウインドウ内で発生するイベントはEventHandlerで処理をおこなう

    メインウィンドウの画像表示部分は3つのレイヤーからなる
      * ペン
      * マスク
      * 原画
    上に書いてあるものを手前に表示している
    マスクの保存は、マスクとペンを合成後におこなう
    葉を塗りわける前にペンとマスクを統合する必要がある

    ショートカットキー
      * Control-O: RGB画像を開く
      * Control-S: マスクを保存
      * Control-Z: アンドゥ
      * Control-Shift-Z: リドゥ
      * Control-E: マスクとペンレイヤーを統合
      * Control + マウスホイール: ズーム
      * Control + ドラッグ: パン
    '''

    def __init__(self, parents=None):
        super().__init__(parents)
        self.initUI()
        self.initHandler()

    def initUI(self):
        '''UIの初期化'''
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        scene = QGraphicsScene(self.ui.graphicsView)
        self.scene = scene
        self.ui.graphicsView.setScene(scene)
        # 原画像
        srcImage = QPixmap(500, 500)
        srcImage.fill(QColor(0, 0, 0))
        self.srcItem = QGraphicsPixmapItem(srcImage)
        self.ui.graphicsView.resize(srcImage.size())
        scene.addItem(self.srcItem)
        # 前フレーム画像
        self.previousItem = QGraphicsPixmapItem(srcImage.copy())
        self.previousItem.setOpacity(0)
        scene.addItem(self.previousItem)
        # マスク画像
        self.maskItem = QGraphicsPixmapItem(srcImage.copy())
        self.maskItem.setOpacity(0.5)
        scene.addItem(self.maskItem)
        #
        self.penItem = QGraphicsPixmapItem(srcImage.copy())
        self.penItem.setOpacity(0.5)
        scene.addItem(self.penItem)
        self.initColorComboBox()

    def initColorComboBox(self):
        '''
        色コンボボックス初期化
        色コンボボックスはInstanceDivider.pyと同じディレクトリ上にある
        leaf.csvより読み込む。フォーマットは
          名前,赤の値,緑の値,青の値
        となっている
        '''
        def add(name, r, g, b, lineno):
            messagePrinted = False
            try:
                pixmap = QPixmap(40, 40)
                color = QColor(int(r), int(g), int(b))
                pixmap.fill(color)
                self.ui.colorComboBox.addItem(QIcon(pixmap), name, color)
            except Exception as exception:
                if not messagePrinted:
                    print('leaf.csvにエラーがあります')
                    messagePrinted = True
                print('%s: %s' % (lineno, exception))
                #raise exception
        dirname = os.path.dirname(sys.argv[0])
        with codecs.open(os.path.join(dirname, 'leaf.csv'), 'r', 'utf-8') as f:
            lineno = 0
            for line in csv.reader(f):
                lineno = lineno + 1
                add(*line, lineno)
            
    def initHandler(self):
        '''イベントハンドラの作成と初期化'''
        handler = self.handler = EventHandler(self)
        ui = self.ui
        # 同時に動かす, 有効 チェックボックス
        # XXX: LayoutBox単位のenable/disableが欲しい…
        ui.sliderSync.clicked.connect(handler.setSliderSync)
        handler.enableSlidersChanged.connect(ui.sliderSync.setEnabled)
        ui.enableSliders.clicked.connect(handler.setEnableSliders)
        handler.enableSlidersChanged.connect(ui.enableSliders.setChecked)
        handler.enableSlidersChanged.connect(ui.redValue.setEnabled)
        handler.enableSlidersChanged.connect(ui.greenValue.setEnabled)
        handler.enableSlidersChanged.connect(ui.blueValue.setEnabled)
        handler.enableSlidersChanged.connect(ui.redSlider.setEnabled)
        handler.enableSlidersChanged.connect(ui.greenSlider.setEnabled)
        handler.enableSlidersChanged.connect(ui.blueSlider.setEnabled)
        # RGB スピンボックス
        # スライダーとスピンボックスはDesigner上で接続済み
        ui.redValue.valueChanged.connect(handler.setValue)
        ui.greenValue.valueChanged.connect(handler.setValue)
        ui.blueValue.valueChanged.connect(handler.setValue)
        handler.redValueChanged.connect(ui.redValue.setValue)
        handler.greenValueChanged.connect(ui.greenValue.setValue)
        handler.blueValueChanged.connect(ui.blueValue.setValue)
        # maskItem
        handler.maskOpacityChangedFloat.connect(self.maskItem.setOpacity)
        # showPreviousFrame
        #ui.showPreviousFrame.clicked.connect(handler.set
        handler.hasPreviousFrameChanged.connect(ui.showPreviousFrame.setEnabled)
        ui.showPreviousFrame.clicked.connect(handler.showPreviousFrame)
        # マスクのα スライダー
        ui.maskOpacitySlider.valueChanged.connect(handler.setMaskOpacityInt)
        handler.maskOpacityChangedInt.connect(ui.maskOpacitySlider.setValue)
        # scene, pen
        self.scene.mousePressEvent = handler.mousePressEvent
        self.scene.mouseMoveEvent = handler.mouseMoveEvent
        self.scene.mouseReleaseEvent = handler.mouseReleaseEvent
        self.scene.keyPressEvent = handler.keyPressEvent
        self.scene.keyReleaseEvent = handler.keyReleaseEvent
        self.scene.wheelEvent = handler.wheelEvent
        self.penItem.paint = handler.paint
        # penSize
        ui.penSizeValue.valueChanged.connect(handler.setPenSize)
        handler.penSizeChanged.connect(ui.penSizeValue.setValue)
        # 色コンボボックス
        ui.colorComboBox.currentIndexChanged.connect(handler.setPenColor)
        # 保存,結合ボタン
        ui.saveButton.clicked.connect(handler.save)
        ui.mergeButton.clicked.connect(handler.merge)
        # メニュー
        self.ui.actionOpen.triggered.connect(handler.open)
        self.ui.actionSave.triggered.connect(handler.save)
        self.ui.actionUndo.triggered.connect(handler.undo)
        self.ui.actionRedo.triggered.connect(handler.redo)
        self.ui.actionMerge.triggered.connect(handler.merge)

    def getSrcPixmap(self):
        '''原画像のQPixmapを返す'''
        return self.srcItem.pixmap()

    def getMaskPixmap(self):
        '''マスク画像のQPixmapを返す'''
        return self.maskItem.pixmap()

    def setMaskPixmap(self, mask):
        '''マスク画像を置き換える'''
        self.maskItem.setPixmap(mask)

    def getPreviousPixmap(self):
        '''前フレームのQPixmapを返す'''
        return self.previousItem.pixmap()

    def setPreviousPixmap(self, previous):
        self.previousItem.setPixmap(previous)

    def getSrcItem(self):
        return self.srcItem

    def getMaskItem(self):
        return self.maskItem

    def getPreviousItem(self):
        return self.previousItem

    def getPenItem(self):
        return self.penItem

    initialized = False
    def resizeEvent(self, event):
        '''
        リサイズイベント
        起動時は、EventHandlerが持つモデルをビューに設定するためにhandlerのresizeEventを実行する
        '''
        if not self.initialized:
            self.handler.resizeEvent(event)
        self.initialized = True
