# coding: utf-8

import numpy as np
import cv2
from PyQt5.QtGui import QPixmap, QImage, QColor
from PyQt5.QtCore import QPoint


'''QPixmapとOpenCV形式画像(以下mat)の相互変換

本システムではQtをメインとし、一部の画像処理にOpenCVを用いる
そのため、画像形式の変換フローは、QPixmap->mat->QPixmapのみとなる
'''


def matFromPixmap(pixmap):
    '''QPixmapからmatに変換'''
    image = pixmap.toImage()
    width = image.width()
    height = image.height()
    depth = int(image.depth() / 8)
    bits = image.constBits().asarray(width * height * depth)
    return np.array(bits).reshape((height, width, depth))


def pixmapFromMat(mat):
    '''matからQPixmapに変換'''
    format = None
    if len(mat.shape) == 2:
        format = QImage.Format_Grayscale8
    elif mat.shape[2] == 3:
        format = QImage.Format_RGB888
    elif mat.shape[2] == 4:
        format = QImage.Format_RGB32
    assert format is not None
    image = QImage(mat.data, mat.shape[1], mat.shape[0], mat.strides[0], format)
    return QPixmap.fromImage(image)


def decompose(pixmap):
    '''
    QPixmapをmatの各チャネルに分解

    戻り値は(R, G, B)並び
    '''
    mats = cv2.split(matFromPixmap(pixmap))
    return [mats[2], mats[1], mats[0]]


def compose(mats):
    '''
    matの各チャネルを合成してQPixmapに変換

    matsは(R, G, B)並びとなっていること
    '''
    return pixmapFromMat(cv2.merge((mats[2], mats[1], mats[0])))


def pixmapPixel(pixmap, point):
    '''pixmapのpoint位置の画素を返す'''
    return pixmap.toImage().pixel(point) & 0xffffff


def matPointFromQPoint(qPoint):
    return (qPoint.x(), qPoint.y())


def qPointFromMATPoint(matPoint):
    return QPoint(matPoint[1], matPoint[0])


def matColorFromQColor(qColor):
    return (qColor.red(), qColor.green(), qColor.blue())


def qColorFromMATColor(matColor):
    return QColor(matColor[0], matColor[1], matColor[2])


def floodFill(pixmap, qPoint, qColor):
    '''pixmap上のqPoint閉領域をqColor塗りつぶす。結果をpixmapにして返す'''
    size = pixmap.size()
    if qPoint.x() < 0 or qPoint.y() < 0:
        return None
    if size.width() <= qPoint.x() or size.height() <= qPoint.y():
        return None
    mat = matFromPixmap(pixmap)
    mat = cv2.split(mat)
    mat = cv2.merge((mat[2], mat[1], mat[0]))
    assert len(mat.shape) == 3 and mat.shape[2] == 3
    pos = matPointFromQPoint(qPoint)
    color = matColorFromQColor(qColor)
    retval, mat, mask, rect = cv2.floodFill(mat, None, pos, color) #, loDiff=0, upDiff=0, flags=0)
    return pixmapFromMat(mat)


def bitwiseAnd(pixmap1, pixmap2):
    mat1 = matFromPixmap(pixmap1)
    mat2 = matFromPixmap(pixmap2)
    return pixmapFromMat(cv2.bitwise_or(mat1, mat2))


def colorHistogram(pixmap):
    '''色の数を返す'''
    histogram = dict()
    image = pixmap.toImage()
    for y in range(image.height()):
        for x in range(image.width()):
            c = image.pixel(x, y)
            histogram[c] = histogram.get(c, 0) + 1
    return histogram


def printColorHistogram(histogram):
    for key in histogram:
        print('%s: %s' % (hex(key), histogram[key]))
